#!/usr/bin/python3
"""depthFirst.py
 This file implements a simple "SOLUZION" automatic solver
that uses depth-first search to try to reach a goal state.
The main purpose of the program is to serve as a reference
algorithm and implementation, in order to try out other
algorithms that take problems in the same format.

S. Tanimoto, April 4, 2017.
"""

# First, get the PROBLEM name from the command-line arguments

import sys, importlib.util

if len(sys.argv)<2:
  print('''
       Usage: 
./depthFirst.py <PROBLEM NAME>
       For example:
./depthFirst.py Missionaries
  ''')
  exit(1)
  
problem_name = sys.argv[1]
print("problem_name = "+problem_name)

# Second, import the problem formulation file itself.

try:
  spec = importlib.util.spec_from_file_location(problem_name, problem_name+".py")
  PROBLEM = spec.loader.load_module()
  spec.loader.exec_module(PROBLEM)
except Exception as e:
  print(e)
  exit(1)


TITLE="Depth-First Solver (Version 0.1)"

def setup():
    print(TITLE)
    print(PROBLEM.PROBLEM_NAME+"; "+PROBLEM.PROBLEM_VERSION)

def report_progress():
    global CURRENT_STATE, OPEN
    print("CURRENT_STATE = "+str(CURRENT_STATE))
    print("OPEN = [",end='')
    for s in OPEN: print('   '+str(s))
    print("]")

# Here are some symbols representing data structures used in
# the search algorithm.
# We make them global so it's easy to call helper functions
# such as report_progress, without having to pass them as arguments.
OPEN = None; CLOSED = None; CURRENT_STATE = None; PROVENANCE = None

# Here is the main depth-first-search algorithm, which is based on
# the pseudocode shown in the CSE 190D second lecture, slide 37.
def dfs():
    global OPEN, CLOSED, CURRENT_STATE, PROVENANCE
    OPEN = [PROBLEM.INITIAL_STATE]
    CLOSED = []
    PROVENANCE = {PROBLEM.hash(PROBLEM.INITIAL_STATE): (None, -1)} # No parent, no operator.
    while len(OPEN)>0:
      S = OPEN.pop(0)    # S is the symbol in the pseudocode. Remove it from OPEN.
      CURRENT_STATE = S  # share S globally as CURRENT_STATE.
      CLOSED.append(S)   # Put it on CLOSED
      # report_progress()
      # Test S to see if it is a goal state.
      if PROBLEM.goal_test(S):
          print('SUCCESS! The search has reached a goal state.')
          report_path()
          print('Do you wish to continue exploring?')
          answer = input("Y or N? >> ")
          if answer=="Y" or answer=="y": print("OK, continue")
          else: return

      L = [] # Let's create the list of successors of S.
      for i in range(len(PROBLEM.OPERATORS)):
          op = PROBLEM.OPERATORS[i]
          if not op.is_applicable(S): continue
          SUCC = op.state_transf(S) ; # print("SUCC = "+str(SUCC))
          if stateIsContainedIn(SUCC, CLOSED): continue
          if stateIsContainedIn(SUCC, OPEN): continue
          PROVENANCE[PROBLEM.hash(SUCC)] = (S, i)
          L.append(SUCC); # print("L = " + str(L))
      # Now insert all members of L at the FRONT of OPEN:
      OPEN = L + OPEN[:]
      # report_progress()
    print("Done")

def report_path():
    global CURRENT_STATE
    path = [CURRENT_STATE]
    operator_sequence = []
    s = CURRENT_STATE
    parent, op_index = PROVENANCE[PROBLEM.hash(s)]
    while op_index > -1:
      path.append(parent)
      operator_sequence.append(op_index)
      parent, op_index = PROVENANCE[PROBLEM.hash(parent)]
    path.reverse()
    operator_sequence.reverse()
    print("Path, as sequence of states:")
    for s in path: print("    "+str(s))
    print("Operator sequence: " + str(operator_sequence))
    return operator_sequence

# The following function determines whether the state given
# is already an element of the list given by argument lst.
# It is important that the comparison of states be perfomed by
# the special method state_equals, which is defined in the problem
# formulation for states of this problem.
def stateIsContainedIn(state, lst):
  for s in lst:
    if PROBLEM.state_equals(state, s): return True
  return False


# The following is only executed if this module is being run as the main
# program, rather than imported from another one.
if __name__ == '__main__':
  setup()
  dfs()
  print("The session is finished.")
